package at.ac.tuwien.iap.imagej;

import ij.*;
import ij.process.*;
import ij.io.*;
import java.awt.image.*;
import java.io.*;

/*
 * This code originally comes from ImageJ itself and is released into the
 * public domain with no specific license.
 * 
 * Micheal Schmid has confirmed that his additions can be released under
 * the same terms as ImageJ
 */

/** ImageJ Plugin for reading an AVI file into an image stack (one slice per video frame) */
/*  By Daniel Marsh
 *  2008-04-11 support for YUYV and YUY2 compression added
 *             uneven sizes and audio fixed
 *             Converts to grayscale with ALT key pressed on start (Michael Schmid)
 *  
 *  2008-04-18 modified to allow operation without going through the
 *             ImageJ GUI.  Changed the method to support returning
 *             an ImagePlus object directly (Simon Andrews)
 */
public class AVI_Reader {
	
	private ImagePlus imagePlus;

    private final static int   YVYU_COMPRESSION = 0x59565955; //'YVYU'
    private final static int   YUY2_COMPRESSION = 0x32595559; //' YUV'
    private  long              startTime;
    private  RandomAccessFile  raFile;
    private  boolean           grayMode;        //whether to convert color video to grayscale

    private  File              file;

    private  String            type = "error";
    private  String            fcc = "error";
    private  int               size = -1;

    private  boolean           verbose = IJ.debugMode;
    private  boolean           showTimes = false;


    //From AVI Header Chunk

    private  int               dwMicroSecPerFrame;
    private  int               dwMaxBytesPerSec;
    private  int               dwReserved1;
    private  int               dwFlags;
    private  int               dwTotalFrames;
    private  int               dwInitialFrames;
    private  int               dwStreams;
    private  int               dwSuggestedBufferSize;
    private  int               dwWidth;
    private  int               dwHeight;
    private  int               dwScale;
    private  int               dwRate;
    private  int               dwStart;
    private  int               dwLength;

    //From Stream Header Chunk

    private  String            fccStreamType;
    private  String            fccStreamHandler;
    private  int               dwStreamFlags;
    private  int               dwStreamReserved1;
    private  int               dwStreamInitialFrames;
    private  int               dwStreamScale;
    private  int               dwStreamRate;
    private  int               dwStreamStart;
    private  int               dwStreamLength;
    private  int               dwStreamSuggestedBufferSize;
    private  int               dwStreamQuality;
    private  int               dwStreamSampleSize;

    //From Stream Format Chunk
    //BMP header reader from BMP class

    // Actual contents (40 bytes):
    private  int               BMPsize;// size of this header in bytes
    private  short             BMPplanes;// no. of color planes: always 1
    private  int               BMPsizeOfBitmap;// size of bitmap in bytes (may be 0: if so, calculate)
    private  int               BMPhorzResolution;// horizontal resolution, pixels/meter (may be 0)
    private  int               BMPvertResolution;// vertical resolution, pixels/meter (may be 0)
    private  int               BMPcolorsUsed;// no. of colors in palette (if 0, calculate)
    private  int               BMPcolorsImportant;// no. of important colors (appear first in palette) (0 means all are important)
    private  int               BMPnoOfPixels;

    private  int               BMPwidth;
    private  int               BMPheight;
    private  short             BMPbitsPerPixel;
    private  int               BMPcompression;

    private  int               BMPactualSizeOfBitmap;
    private  int               BMPscanLineSize;
    private  int               BMPactualColorsUsed;

    private  int[]             intData;
    private  byte[]            byteData;
    private  byte[]            rawData;

    //palette
    private  byte[]            r;
    private  byte[]            g;
    private  byte[]            b;

    private  ColorModel        cm;
    private  ImageProcessor    ip;

    private  ImageStack        stack = new ImageStack(0, 0);

    private  long              pos;

    private  long              bigChunkSize = 0;

    private  String            fileName;

    private  long              lastTime = 0;


    public ImagePlus loadAVI (String arg) {
        grayMode = IJ.altKeyDown();
        imagePlus = null;
        try {
            IJ.showProgress(.01);
            readAVI(arg);
        } catch (OutOfMemoryError e) {
            //IJ.outOfMemory(fileName);
            stack.trim();
            IJ.showMessage("AVI Reader", "Out of memory.  " + stack.getSize() + " of " + dwTotalFrames + " frames will be opened.");
            if (stack != null && stack.getSize() > 0) {
                imagePlus =  new ImagePlus(fileName, stack);
            }
            //break; // break out of the frame reading loop
        } catch (Exception e) {
            String  msg  = e.getMessage();
            if (msg == null || msg.equals("")) {
                msg = "" + e;
            }

            IJ.showMessage("AVI Reader", "An error occurred reading the file.\n \n" + msg);
        } finally {
            try {
                showTime("Closing File");
                raFile.close();
                if (verbose)
                    IJ.log("File closed.");
            } catch (Exception e) {

            }
            IJ.showProgress(1);
            long  totalTime  = (System.currentTimeMillis() - startTime);
            if (totalTime==0)
                totalTime = 1;
            IJ.showStatus("Done in " + totalTime + " msec.  " + (stack.getWidth() * stack.getHeight() * stack.getSize() / totalTime * 1000) + " pixels/second");
        }
        return imagePlus;
    }


    public void readAVI(String path) throws Exception, IOException {

        if (showTimes)
            IJ.write("/--------------------------/");

        //lutBufferRemapped = new int[1];

        OpenDialog  sd = new OpenDialog("Select AVI File", path);
        startTime = System.currentTimeMillis();
        showTime("File selected");
        fileName    = sd.getFileName();
        if (fileName == null)
            return;
        String      fileDir     = sd.getDirectory();
        file = new File(fileDir + fileName);
        raFile = new RandomAccessFile(file, "r");

        if (verbose)
            IJ.log("/--------------------------/");

        showTime("Reading File Header");
        readFileHeader();

        byte[]      list        = new byte[4];
        String      listString;

        //IJ.write("bigChunkSize="+bigChunkSize);

        while (raFile.read(list) == 4) {
            raFile.seek(raFile.getFilePointer() - 4);

            listString = new String(list);

            updateProgress();

            if (listString.equals("JUNK")) {
                showTime("Skipping JUNK");
                skipJUNK();
            } else if (listString.equals("LIST")) {
                readTypeAndSizeAndFcc();
                raFile.seek(raFile.getFilePointer() - 12);
                if (fcc.equals("hdrl")) {
                    readAviHeader();
                } else if (fcc.equals("strl")) {
                    showTime("Reading Video Stream");
                    long  startPos    = raFile.getFilePointer();
                    long  streamSize  = size;
                    readVideoStream();
                    raFile.seek(startPos + 8 + streamSize);
                } else if (fcc.equals("movi")) {
                    showTime("Reading Movie Data");
                    readMovieData();
                    showTime("Creating ImagePlus to display stack");
                    if (stack != null && stack.getSize() > 0) {
                        imagePlus = new ImagePlus(fileName, stack);
                    }
                    return;
                } else {
                    showTime("Skipping unknown block");
                    if (verbose)
                        IJ.log("**Unsupported LIST fcc='" + fcc + "' detected at 0x" +
                                Long.toHexString(raFile.getFilePointer()));
                    raFile.seek(raFile.getFilePointer() + 8 + size);
                    if (verbose)
                        IJ.log("  Skipped "+size+" bytes, now at 0x" + Long.toHexString(raFile.getFilePointer()));
                }
            } else {
                showTime("Skipping unknown block");
                if (verbose)
                    IJ.log("**Unsupported type '" + listString + "' detected at 0x" +
                            Long.toHexString(raFile.getFilePointer()));
                readTypeAndSize();
                raFile.seek(raFile.getFilePointer() + size);
                if (verbose)
                    IJ.log("  Skipped "+size+" bytes, now at 0x" + Long.toHexString(raFile.getFilePointer()));
            }
        }
        raFile.close();
        return;
    }



    void readFileHeader() throws Exception, IOException {
        readTypeAndSizeAndFcc();
        if (verbose)
            IJ.log("FILE HEADER");
        if (type.equals("RIFF")) {
            if (verbose) {
                IJ.log("RIFF format detected...");
                IJ.log("   size=" + size);
                IJ.log("   fcc=" + fcc);
            }
            bigChunkSize = size;
            if (!fcc.equals("AVI ")) {
                if (verbose) {
                    IJ.log("Unsupported format '" + fcc + "'");
                    IJ.log("Expected format 'AVI '");
                }
                throw new Exception("Unsupported file type.  AVI RIFF form required.");
            }
        } else {
            if (verbose) {
                IJ.log("Unexpected '" + type + "'");
                IJ.log("Expected 'RIFF' not found");
            }
            throw new Exception("The file does not appear to be in AVI format.");
        }
    }


    void readAviHeader() throws IOException {
        if (verbose)
            IJ.log("AVI HEADER");
        readTypeAndSizeAndFcc();

        if (type.equals("LIST")) {
            if (verbose) {
                IJ.log("AVI header detected...");
                IJ.log("   size=" + size);
                IJ.log("   fcc=" + fcc);
            }
            if (fcc.equals("hdrl")) {
                readTypeAndSize();
                if (type.equals("avih")) {
                    if (verbose)
                        IJ.log("   AVI header chunk (avih) detected...");
                    pos = raFile.getFilePointer();

                    dwMicroSecPerFrame = readInt();
                    dwMaxBytesPerSec = readInt();
                    dwReserved1 = readInt();
                    dwFlags = readInt();
                    dwTotalFrames = readInt();
                    dwInitialFrames = readInt();
                    dwStreams = readInt();
                    dwSuggestedBufferSize = readInt();
                    dwWidth = readInt();
                    dwHeight = readInt();
                    dwScale = readInt();
                    dwRate = readInt();
                    dwStart = readInt();
                    dwLength = readInt();

                    if (verbose) {
                        IJ.log("      dwMicroSecPerFrame=" + dwMicroSecPerFrame);
                        IJ.log("      dwMaxBytesPerSec=" + dwMaxBytesPerSec);
                        IJ.log("      dwReserved1=" + dwReserved1);
                        IJ.log("      dwFlags=" + dwFlags);
                        IJ.log("      dwTotalFrames=" + dwTotalFrames);
                        IJ.log("      dwInitialFrames=" + dwInitialFrames);
                        IJ.log("      dwStreams=" + dwStreams);
                        IJ.log("      dwSuggestedBufferSize=" + dwSuggestedBufferSize);
                        IJ.log("      dwWidth=" + dwWidth);
                        IJ.log("      dwHeight=" + dwHeight);
                        IJ.log("      dwScale=" + dwScale);
                        IJ.log("      dwRate=" + dwRate);
                        IJ.log("      dwStart=" + dwStart);
                        IJ.log("      dwLength=" + dwLength);
                    }

                    raFile.seek(pos + size);
                }
            } else {
                IJ.log("**Unsupported fcc '" + fcc + "'; expected fcc 'hdrl'");
                return;
            }
        } else {
            IJ.log("**Unexpected '" + type + "'; expected 'LIST' not found");
            return;
        }
    }



    void readVideoStream() throws Exception, IOException {
        if (verbose)
            IJ.log("VIDEO STREAM HEADER");
        readTypeAndSizeAndFcc();

        if (type.equals("LIST")) {
            if (fcc.equals("strl")) {
                if (verbose) {
                    IJ.log("Stream header (strl) detected...");
                    IJ.log("   size=" + size);
                    IJ.log("   fcc=" + fcc);
                }

                readTypeAndSize();
                if (type.equals("strh")) {
                    if (verbose)
                        IJ.log("   Stream header chunk (strh) detected...");

                    pos = raFile.getFilePointer();

                    String  fccStreamTypeOld;
                    fccStreamTypeOld = fccStreamType;
                    fccStreamType = readStringBytes();

                    if (!fccStreamType.equals("vids")) {
                        if (verbose)
                            IJ.log("      Not video stream (fcc '" + fccStreamType + "')");

                        fccStreamType = fccStreamTypeOld;

                        return;
                    }

                    readStrh();

                    raFile.seek(pos + size);
                } else {
                    IJ.log("**Expected fcc 'strh', found fcc '" + fcc + "' @ 0x"+Long.toHexString(raFile.getFilePointer()-4));
                    return;
                }
                readTypeAndSize();
                if (type.equals("strf")) {
                    if (verbose)
                        IJ.log("   Stream format chunk (strf) detected...");

                    pos = raFile.getFilePointer();
                    readStrf();
                    raFile.seek(pos + size);
                } else {
                    IJ.log("**Expected fcc 'strf', found fcc '" + fcc + "' @ 0x"+Long.toHexString(raFile.getFilePointer()-4));
                    return;
                }
            } else {
                IJ.log("**Expected fcc 'strl', found fcc '" + fcc + "' @ 0x"+Long.toHexString(raFile.getFilePointer()-4));
                return;
            }

            readTypeAndSize();
            if (type.equals("strd")) {
                if (verbose)
                    IJ.log("   Stream 'strd' chunk detected and skipped ("+size+")");
                raFile.seek(raFile.getFilePointer() + size);
            } else {
                if (verbose)
                    IJ.log("   Type '" + type + "' detected.  Backing up to 0x"+Long.toHexString(raFile.getFilePointer()-8));
                raFile.seek(raFile.getFilePointer() - 8);
            }
            readTypeAndSize();
            if (type.equals("strn")) {
                if (verbose)
                    IJ.log("   Stream 'strn' chunk detected and skipped ("+size+" bytes)");
                raFile.seek(raFile.getFilePointer() + size);
            } else {
                if (verbose)
                    IJ.log("   Type '" + type + "' detected.  Backing up to 0x"+Long.toHexString(raFile.getFilePointer()-8));
                raFile.seek(raFile.getFilePointer() - 8);
            }
        } else {
            IJ.log("**Unexpected '" + type + "'; expected 'LIST' not found");
            return;
        }

        if (verbose)
            IJ.log("");
    }


    void readStrh() throws IOException {
        fccStreamHandler = readStringBytes();
        dwStreamFlags = readInt();
        dwStreamReserved1 = readInt();
        dwStreamInitialFrames = readInt();
        dwStreamScale = readInt();
        dwStreamRate = readInt();
        dwStreamStart = readInt();
        dwStreamLength = readInt();
        dwStreamSuggestedBufferSize = readInt();
        dwStreamQuality = readInt();
        dwStreamSampleSize = readInt();
        if (verbose) {
            IJ.log("      fccStreamType=" + fccStreamType);
            IJ.log("      fccStreamHandler=" + fccStreamHandler);
            IJ.log("      dwStreamFlags=" + dwStreamFlags);
            IJ.log("      dwStreamReserved1=" + dwStreamReserved1);
            IJ.log("      dwStreamInitialFrames=" + dwStreamInitialFrames);
            IJ.log("      dwStreamScale=" + dwStreamScale);
            IJ.log("      dwStreamRate=" + dwStreamRate);
            IJ.log("      dwStreamStart=" + dwStreamStart);
            IJ.log("      dwStreamLength=" + dwStreamLength);
            IJ.log("      dwStreamSuggestedBufferSize=" + dwStreamSuggestedBufferSize);
            IJ.log("      dwStreamQuality=" + dwStreamQuality);
            IJ.log("      dwStreamSampleSize=" + dwStreamSampleSize);
        }
    }


    void readStrf() throws Exception, IOException {
        //BMP header reader from BMP class

        BMPsize = readInt();
        BMPwidth = readInt();
        BMPheight = readInt();
        BMPplanes = readShort();
        BMPbitsPerPixel = readShort();
        BMPcompression = readInt();
        BMPsizeOfBitmap = readInt();
        BMPhorzResolution = readInt();
        BMPvertResolution = readInt();
        BMPcolorsUsed = readInt();
        BMPcolorsImportant = readInt();

        BMPnoOfPixels = BMPwidth * BMPheight;

        // Scan line is padded with zeroes to be a multiple of four bytes
        BMPscanLineSize = ((BMPwidth * BMPbitsPerPixel + 31) / 32) * 4;

        if (BMPsizeOfBitmap != 0) {
            BMPactualSizeOfBitmap = BMPsizeOfBitmap;
        } else {
            // a value of 0 doesn't mean zero - it means we have to calculate it
            BMPactualSizeOfBitmap = BMPscanLineSize * BMPheight;
        }

        if (BMPcolorsUsed != 0) {
            BMPactualColorsUsed = BMPcolorsUsed;
        } else
        // a value of 0 means we determine this based on the bits per pixel
        if (BMPbitsPerPixel < 16) {
            BMPactualColorsUsed = 1 << BMPbitsPerPixel;
            BMPcolorsUsed = BMPactualColorsUsed; //Add this line
        } else {
            BMPactualColorsUsed = 0;
        }            // no palette

        if (verbose) {
            IJ.log("      BMPsize=" + BMPsize);
            IJ.log("      BMPwidth=" + BMPwidth);
            IJ.log("      BMPheight=" + BMPheight);
            IJ.log("      BMPplanes=" + BMPplanes);
            IJ.log("      BMPbitsPerPixel=" + BMPbitsPerPixel);
            IJ.log("      BMPcompression=" + BMPcompression);
            IJ.log("      BMPsizeOfBitmap=" + BMPsizeOfBitmap);
            IJ.log("      BMPhorzResolution=" + BMPhorzResolution);
            IJ.log("      BMPvertResolution=" + BMPvertResolution);
            IJ.log("      BMPcolorsUsed=" + BMPcolorsUsed);
            IJ.log("      BMPcolorsImportant=" + BMPcolorsImportant);
            IJ.log("      >BMPnoOfPixels=" + BMPnoOfPixels);
            IJ.log("      >BMPscanLineSize=" + BMPscanLineSize);
            IJ.log("      >BMPactualSizeOfBitmap=" + BMPactualSizeOfBitmap);
            IJ.log("      >BMPactualColorsUsed=" + BMPscanLineSize);

            IJ.log("      Read up to " + raFile.getFilePointer());
            IJ.log("      Format ends at " + (pos + size));
        }

        if (BMPcompression != 0 && BMPcompression != YVYU_COMPRESSION && BMPcompression != YUY2_COMPRESSION) {
            if (verbose)
                IJ.log("**Unsupported compression: 0x"+Integer.toHexString(BMPcompression));
            throw new Exception("Unsupported compression: 0x"+Integer.toHexString(BMPcompression));
        }

        if (!(BMPbitsPerPixel==8 || BMPbitsPerPixel==24 || BMPbitsPerPixel==32 ||
                (BMPcompression!=0 && BMPbitsPerPixel==16))) {    //16-bit-per pixel compressed files are supported
            if (verbose)
                IJ.log("**Unsupported bits per pixel: "+BMPbitsPerPixel);
            throw new Exception("Unsupported bits-per-pixel value: "+BMPbitsPerPixel);
        }

        if (BMPactualColorsUsed != 0) {
            if (verbose)
                IJ.log("      Now reading palette..."); // + (pos + size) );

            long    pos1  = raFile.getFilePointer();

            byte[]  pr    = new byte[BMPcolorsUsed];
            byte[]  pg    = new byte[BMPcolorsUsed];
            byte[]  pb    = new byte[BMPcolorsUsed];
            for (int i = 0; i < BMPcolorsUsed; i++) {
                pb[i] = raFile.readByte();
                pg[i] = raFile.readByte();
                pr[i] = raFile.readByte();
                raFile.readByte();
            }
            if (verbose) {
                IJ.log("      Palette was " + (raFile.getFilePointer() - pos1) + " bytes long");
                IJ.log("      Palette ended at " + (raFile.getFilePointer()) + " and was expected to end at " + (pos + size));
            }
            cm = new IndexColorModel(BMPbitsPerPixel, BMPcolorsUsed, pr, pg, pb);

            if (verbose)
                IJ.log("      Creating stack with palette...");
            stack = new ImageStack(dwWidth, BMPheight, cm);

        } else {
            if (verbose)
                IJ.log("      Creating stack...");
            stack = new ImageStack(dwWidth, BMPheight);
        }
    }


    void skipJUNK() throws IOException {
        readTypeAndSize();
        if (type.equals("JUNK")) {
            if (verbose) 
                IJ.log("JUNK stream detected and skipped ("+size+" bytes)");
            raFile.seek(raFile.getFilePointer() + size);
        }
    }


    void checkList(String thefcc) throws IOException {
        readTypeAndSizeAndFcc();
        if (!fcc.equals(thefcc)) {
            if (verbose) 
                IJ.log("Type '" + type + "' with fcc '" + fcc + "' detected and skipped ("+size+" bytes)");
            raFile.seek(raFile.getFilePointer() + size);
        } else 
            raFile.seek(raFile.getFilePointer() - 12);
        if (verbose)
            IJ.log("");
    }


    void readMovieData() throws Exception, IOException {
        if (verbose)
            IJ.log("STREAM DATA");
        readTypeAndSizeAndFcc();

        if (type.equals("LIST")) {
            if (verbose)
                IJ.log("Type 'LIST' detected of fcc '" + fcc + "'...");
            if (fcc.equals("movi")) {
                if (verbose)
                    IJ.log("Movie data detected...");
                readTypeAndSizeAndFcc();
                if (type.equals("LIST") && fcc.equals("rec ")) {
                    if (verbose)
                        IJ.log("   Movie record detected and skipped");
                } else {
                    if (verbose)
                        IJ.log("   Type '" + type + "' and fcc '" + fcc + "' detected.  Backing up to 0x"+
                                Long.toHexString(raFile.getFilePointer()-12));
                    raFile.seek(raFile.getFilePointer() - 12);
                }
                readTypeAndSize();
                long  startPos  = raFile.getFilePointer();
                if (BMPbitsPerPixel <= 8 || grayMode)
                    ip = new ByteProcessor(dwWidth, BMPheight);
                else
                    ip = new ColorProcessor(dwWidth, BMPheight);

                showTime(">>Entering while-loop to read chunks");
                while (type.substring(2).equals("db") || type.substring(2).equals("dc") || type.substring(2).equals("wb")) {
                    updateProgress();
                    pos = raFile.getFilePointer();
                    if (type.substring(2).equals("db") || type.substring(2).equals("dc")) { //video chunks only
                        if (verbose)
                            IJ.log("   Video data chunk (" + type + ") detected; size="+size+" bytes");
                        readFrame();
                        if (BMPbitsPerPixel <= 8 || grayMode)
                            ip.setPixels(byteData);
                        else
                            ip.setPixels(intData);
                        if (BMPactualColorsUsed != 0) {
                            ip.setColorModel(cm);
                            if (verbose)
                                IJ.log("      Color model set");
                        }
                        stack.addSlice("", ip);

                    } else {    //non-video chunk
                        raFile.seek(raFile.getFilePointer() + size);
                        if (verbose)
                            IJ.log("   Non-video data chunk '" + type + "', skipped ("+size+" bytes)");
                    }
                    readTypeAndSize();
                    if (type.equals("JUNK")) {
                        raFile.seek(raFile.getFilePointer() + size);
                        readTypeAndSize();
                    }
                }

                showTime(">>Exiting while-loop to read chunks");
                if (verbose)
                    IJ.log("End of video data reached with type '" + type + "'.  Backing up.");
                raFile.seek(raFile.getFilePointer() - 8);
            } else 
                IJ.log("**Expected fcc 'movi', but found fcc '" + fcc + "'");
        } else
            IJ.log("**Expected 'LIST', but found '" + type + "'");
        if (verbose) 
            IJ.log("");
    }

    /** For one line: copy byte data into the byte array for creating a ByteProcessor */
    void unpack8bit(byte[] rawData, int rawOffset, byte[] byteData, int byteOffset, int w) {
        for (int i = 0; i < w; i++)
            byteData[byteOffset + i] = rawData[rawOffset + i];
    }

    /** For one line: Unpack and convert YUV or RGB video data to grayscale (byte array for ByteProcessor) */
    void unpackGray(byte[] rawData, int rawOffset, byte[] byteData, int byteOffset, int w) {
        int  j     = byteOffset;
        int  k     = rawOffset;
        if (BMPcompression == 0) {
            for (int i = 0; i < w; i++) {
                int  b0  = (((int) (rawData[k++])) & 0xff);
                int  b1  = (((int) (rawData[k++])) & 0xff);
                int  b2  = (((int) (rawData[k++])) & 0xff);
                if (BMPbitsPerPixel==32) k++; // ignore 4th byte (alpha value) 
                byteData[j++] = (byte)((b0*934 + b1*4809 + b2*2449 + 4096)>>13); //0.299*R+0.587*G+0.114*B
            }
        } else { // if (BMPcompression==YVYU_COMPRESSION || BMPcompression==YUY2_COMPRESSION) {
            if (BMPcompression==YVYU_COMPRESSION) k++; //skip first u byte
            for (int i = 0; i < w/2; i++) {
                byteData[j++] = rawData[k++];   //y value. NOTE: no scaling here (use B&C if scaling is desired)
                k++;    //skip v byte
                byteData[j++] = rawData[k++];
                k++;    //skip u byte
            }
        }
    }

    /** For one line: Read YUV, RGB or RGB+alpha data and writes RGB int array for ColorProcessor */
    void unpack(byte[] rawData, int rawOffset, int[] intData, int intOffset, int w) {
        int  j     = intOffset;
        int  k     = rawOffset;
        if (BMPcompression == 0) {
            for (int i = 0; i < w; i++) {
                int  b0  = (((int) (rawData[k++])) & 0xff);
                int  b1  = (((int) (rawData[k++])) & 0xff) << 8;
                int  b2  = (((int) (rawData[k++])) & 0xff) << 16;
                if (BMPbitsPerPixel==32) k++; // ignore 4th byte (alpha value) 
                intData[j] = 0xff000000 | b0 | b1 | b2;
                j++;
            }
        } else if (BMPcompression==YVYU_COMPRESSION) {
            for (int i = 0; i < w/2; i++) {
                int v  = rawData[k++] ^ 0xffffff80; //converts byte range 0...ff to -128 ... 127
                int y0 = rawData[k++] & 0xff;
                int u  = rawData[k++] ^ 0xffffff80;
                int y1 = rawData[k++] & 0xff;
                writeRGBfromYUV(y0, u, v, j++);
                writeRGBfromYUV(y1, u, v, j++);
            }
        } else if (BMPcompression==YUY2_COMPRESSION) {
            for (int i = 0; i < w/2; i++) {
                int y0 = rawData[k++] & 0xff;
                int u  = rawData[k++] ^ 0xffffff80;
                int y1 = rawData[k++] & 0xff;
                int v  = rawData[k++] ^ 0xffffff80;
                writeRGBfromYUV(y0, u, v, j++);
                writeRGBfromYUV(y1, u, v, j++);
            }
        }
    }

    /** Write an intData RGB value converted from YUV,
     *  The y range between 16 and 235 becomes 0...255
     *  u, v should be between -112 and +112
     */
    void writeRGBfromYUV(int y, int u, int v, int intArrayIndex) {
        //int r = (int)(1.164*(y-16)+1.596*v+0.5);
        //int g = (int)(1.164*(y-16)-0.391*u-0.813*v+0.5);
        //int b = (int)(1.164*(y-16)+2.018*u+0.5);
        int r = (9535*y + 13074*v -148464) >> 13;
        int g = (9535*y - 6660*v - 3203*u -148464) >> 13;
        int b = (9535*y + 16531*u -148464) >> 13;
        if (r>255) r=255; if (r<0) r=0;
        if (g>255) g=255; if (g<0) g=0;
        if (b>255) b=255; if (b<0) b=0;
        intData[intArrayIndex] = 0xff000000 | (r<<16) | (g<<8) | b;
    }

    void readFrame() throws Exception, IOException {
        boolean flipY = BMPcompression == 0;    //raw starts at bottom, YUYV etc at top
        int  len = BMPscanLineSize;
        if (BMPbitsPerPixel <=8 || grayMode)
            byteData = new byte[dwWidth * BMPheight];
        else
            intData = new int[dwWidth * BMPheight];
        rawData = new byte[BMPactualSizeOfBitmap];
        int  rawOffset  = 0;
        int  offset     = flipY ? (BMPheight - 1) * dwWidth : 0;
        for (int i = BMPheight - 1; i >= 0; i--) {
            int  n  = raFile.read(rawData, rawOffset, len);
            if (n < len) 
                throw new Exception("Scan line ended prematurely after " + n + " bytes");
            if (BMPbitsPerPixel <=8)
                unpack8bit(rawData, rawOffset, byteData, offset, dwWidth);
            else if (grayMode)
                unpackGray(rawData, rawOffset, byteData, offset, dwWidth);
            else
                unpack(rawData, rawOffset, intData, offset, dwWidth);
            rawOffset += len;
            offset += flipY ? -dwWidth : dwWidth;
        }
    }

    void getPalette() throws IOException {
        int  noOfEntries  = BMPactualColorsUsed;
        //IJ.log("noOfEntries: " + noOfEntries);
        if (noOfEntries > 0) {
            r = new byte[noOfEntries];
            g = new byte[noOfEntries];
            b = new byte[noOfEntries];

            int  reserved;
            for (int i = 0; i < noOfEntries; i++) {
                b[i] = (byte) raFile.read();
                g[i] = (byte) raFile.read();
                r[i] = (byte) raFile.read();
                reserved = raFile.read();
            }
        }
    }


    String readStringBytes() throws IOException {
        //reads next  4 bytes and returns them as a string
        byte[]  list;
        list = new byte[4];
        raFile.read(list);
        return new String(list);
    }



    int readInt() throws IOException {
        // 4 bytes
        // http://mindprod.com/endian.html
        int  accum    = 0;
        int  shiftBy;
        for (shiftBy = 0; shiftBy < 32; shiftBy += 8) {
            accum |= (raFile.readByte() & 0xff) << shiftBy;
        }
        return accum;
    }


    short readShort() throws IOException {
        // 2 bytes
        int  low   = raFile.readByte() & 0xff;
        int  high  = raFile.readByte() & 0xff;
        return (short) (high << 8 | low);
    }

    void readTypeAndSize() throws IOException {//(String listString, int fileLength) throws IOException{
        long pointer = raFile.getFilePointer();
        if (pointer%2==1L) raFile.seek(pointer + 1);    //pad to even address
        type = readStringBytes();
        size = readInt();
        //IJ.log("readTypeAndSize: "+type + " " + size+ " @ 0x"+Long.toHexString(raFile.getFilePointer()-8) );
    }


    void readTypeAndSizeAndFcc() throws IOException {// String listString, int fileLength, String fccString) throws IOException{
        long pointer = raFile.getFilePointer();
        if (pointer%2==1L) raFile.seek(pointer + 1);    //pad to even address
        //listString = readStringBytes();
        //fileLength = readInt();
        //fccString = readStringBytes();
        type = readStringBytes();
        size = readInt();
        fcc = readStringBytes();
    }


    void updateProgress() throws IOException {
        IJ.showProgress((double) raFile.getFilePointer() / bigChunkSize);
        //IJ.write("update");
    }


    void showTime(String w) {
        if (showTimes) {
            long  thisTime  = System.currentTimeMillis();
            if (lastTime > 0) {
                IJ.write("    " + (thisTime - lastTime) + " spent between");
            }
            IJ.write((thisTime - startTime) + ": " + w);
            lastTime = thisTime;
        }
    }

}

